@@ -34,24 +34,12 @@
 #define __LITTLE_ENDIAN	LITTLE_ENDIAN
 
 /* FreeBSD file API is 64-bit aware */
-#define fstat64		fstat
-#define ftruncate64	ftruncate
-#define lseek64		lseek
-#define stat64		stat
-#define pwrite64	pwrite
-#define pread64		pread
 #define fdatasync	fsync
 #define memalign(a,sz)	valloc(sz)
 
 #define EFSCORRUPTED	990	/* Filesystem is corrupted */
 #define EFSBADCRC	991	/* Bad CRC detected */
 
-typedef off_t		xfs_off_t;
-typedef off_t		off64_t;
-typedef __uint64_t	xfs_ino_t;
-typedef __uint32_t	xfs_dev_t;
-typedef __int64_t	xfs_daddr_t;
-
 typedef unsigned char		__u8;
 typedef signed char		__s8;
 typedef unsigned short		__u16;
@@ -61,6 +49,12 @@ typedef signed int		__s32;
 typedef unsigned long long int	__u64;
 typedef signed long long int	__s64;
 
+typedef off_t		xfs_off_t;
+typedef uint64_t	xfs_ino_t;
+typedef uint32_t	xfs_dev_t;
+typedef int64_t		xfs_daddr_t;
+typedef __u32		xfs_nlink_t;
+
 #define	O_LARGEFILE	0
 
 #define HAVE_FID	1
@@ -180,7 +174,8 @@ struct fsxattr {
 	__u32		fsx_extsize;	/* extsize field value (get/set)*/
 	__u32		fsx_nextents;	/* nextents field value (get)	*/
 	__u32		fsx_projid;	/* project identifier (get/set) */
-	unsigned char	fsx_pad[12];
+	__u32		fsx_cowextsize;	/* cow extsize field value (get/set) */
+	unsigned char	fsx_pad[8];
 };
 
 /*
@@ -208,5 +203,8 @@ struct fsxattr {
 
 #endif
 
+#ifndef FS_XFLAG_COWEXTSIZE
+#define FS_XFLAG_COWEXTSIZE	0x00010000	/* CoW extent size allocator hint */
+#endif
 
 #endif	/* __XFS_FREEBSD_H__ */
