@@ -3,8 +3,8 @@
 
 #include <semaphore.h>
 #include "incore.h"
-#include "radix-tree.h"
 
+struct workqueue;
 
 extern int 	do_prefetch;
 
@@ -14,8 +14,7 @@ typedef struct prefetch_args {
 	pthread_mutex_t		lock;
 	pthread_t		queuing_thread;
 	pthread_t		io_threads[PF_THREAD_COUNT];
-	struct radix_tree_root	primary_io_queue;
-	struct radix_tree_root	secondary_io_queue;
+	struct btree_root	*io_queue;
 	pthread_cond_t		start_reading;
 	pthread_cond_t		start_processing;
 	int			agno;
@@ -42,6 +41,15 @@ start_inode_prefetch(
 	int			dirs_only,
 	prefetch_args_t		*prev_args);
 
+void
+do_inode_prefetch(
+	struct xfs_mount	*mp,
+	int			stride,
+	void			(*func)(struct workqueue *,
+					xfs_agnumber_t, void *),
+	bool			check_cache,
+	bool			dirs_only);
+
 void
 wait_for_inode_prefetch(
 	prefetch_args_t		*args);
@@ -52,8 +60,15 @@ cleanup_inode_prefetch(
 
 
 #ifdef XR_PF_TRACE
+void	pftrace_init(void);
+void	pftrace_done(void);
+
 #define pftrace(msg...)	_pftrace(__FUNCTION__, ## msg)
 void	_pftrace(const char *, const char *, ...);
+#else
+static inline void pftrace_init(void) { };
+static inline void pftrace_done(void) { };
+static inline void pftrace(const char *msg, ...) { };
 #endif
 
 #endif /* _XFS_REPAIR_PREFETCH_H */
